/* See LICENSE file for copyright and license details. */
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>

#include <net/if.h>
#ifdef __linux__
#include <linux/if_link.h>
#else
#include <sys/sysctl.h>
#include <net/if_dl.h>
#include <net/route.h>
#endif

#include <err.h>
#include <errno.h>
#include <ifaddrs.h>
#include <limits.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#ifdef __linux__
#define SIGINFO SIGUSR1
#endif

#include "arg.h"

char *argv0;
char *ifname;
uint64_t origrxbytes, origtxbytes, rxbytes, txbytes;
int Bflag;
int tflag;

void
scan(void)
{
	struct ifaddrs *ifas, *ifa;
	int found = 0;

	if (getifaddrs(&ifas) < 0)
		err(1, "getifaddrs");
	if (*ifname) {
		for (ifa = ifas; ifa; ifa = ifa->ifa_next) {
			if (!strncmp(ifname, ifa->ifa_name, IF_NAMESIZE)) {
				found = 1;
				break;
			}
		}
	} else {
		for (ifa = ifas; ifa; ifa = ifa->ifa_next) {
			if (ifa->ifa_flags & IFF_LOOPBACK ||
			    !(ifa->ifa_flags & IFF_UP) ||
			    !(ifa->ifa_flags & IFF_RUNNING))
				continue;
			strncpy(ifname, ifa->ifa_name, IF_NAMESIZE);
			ifname[IF_NAMESIZE - 1] = '\0';
			found = 1;
			break;
		}
	}
	freeifaddrs(ifas);
	if (!found)
		errx(1, "no usable interface found");
}

#ifdef __linux__
void
sample(uint64_t *rxbytes, uint64_t *txbytes,
       uint64_t *rxpps, uint64_t *txpps)
{
	struct ifaddrs *ifas, *ifa;
	struct rtnl_link_stats *stats = NULL;

	if (getifaddrs(&ifas) < 0)
		err(1, "getifaddrs");
	for (ifa = ifas; ifa; ifa = ifa->ifa_next) {
		if (strcmp(ifa->ifa_name, ifname))
			continue;
		if (!ifa->ifa_data || !ifa->ifa_addr)
			break;
		if (ifa->ifa_addr->sa_family == AF_PACKET) {
			stats = ifa->ifa_data;
			*rxbytes = stats->rx_bytes;
			*txbytes = stats->tx_bytes;
			*rxpps = stats->rx_packets;
			*txpps = stats->tx_packets;
			break;
		}
	}
	freeifaddrs(ifas);
	if (!stats)
		errx(1, "interface %s cannot be sampled", ifname);
}
#else
void
sample(uint64_t *rxbytes, uint64_t *txbytes,
       uint64_t *rxpps, uint64_t *txpps)
{
	int mib[6];
	char *buf, *next;
	size_t size;
	struct if_msghdr *ifm;
	struct sockaddr_dl *sdl = NULL;

	*rxbytes = *txbytes = *rxpps = *txpps = 0;
	mib[0] = CTL_NET;
	mib[1] = PF_ROUTE;
	mib[2] = 0;
	mib[3] = 0;
	mib[4] = NET_RT_IFLIST;
	mib[5] = 0;
	if (sysctl(mib, 6, NULL, &size, NULL, 0) < 0)
		errx(1, "sysctl failed");
	buf = malloc(size);
	if (!buf)
		err(1, "malloc");
	if (sysctl(mib, 6, buf, &size, NULL, 0) < 0)
		errx(1, "sysctl failed");
	for (next = buf; next < buf + size; next += ifm->ifm_msglen) {
		ifm = (struct if_msghdr *)next;
		if (ifm->ifm_type == RTM_NEWADDR ||
		    !(ifm->ifm_flags & IFF_UP))
			continue;
		sdl = (struct sockaddr_dl *)(ifm + 1);
		if (sdl->sdl_family != AF_LINK ||
		    strncmp(sdl->sdl_data, ifname, sdl->sdl_nlen))
			continue;
		*rxbytes = ifm->ifm_data.ifi_ibytes;
		*txbytes = ifm->ifm_data.ifi_obytes;
		*rxpps = ifm->ifm_data.ifi_ipackets;
		*txpps = ifm->ifm_data.ifi_opackets;
		break;
	}
	free(buf);
	if (!sdl)
		errx(1, "interface %s cannot be sampled", ifname);
}
#endif

double
scale(char **suffix, uint64_t bits)
{
	static char *suffixes[][5] = { { "b", "kb", "Mb", "Gb", "Tb" },
	                               { "B", "kB", "MB", "GB", "TB" } };
	double rounded = bits;
	int unit;

	if (Bflag)
		rounded /= 8;
	for (unit = 0; rounded >= 1000 && unit < 4; ++unit)
		rounded /= 1000;
	*suffix = suffixes[Bflag][unit];
	return rounded;
}

void
print(uint64_t rxbits, uint64_t txbits, uint64_t rxpps, uint64_t txpps)
{
	char *rxsuffix, *txsuffix;
	double rxround, txround;

	if (!tflag) {
		rxround = scale(&rxsuffix, rxbits);
		txround = scale(&txsuffix, txbits);
		printf("%s: %6.2f %2sps Rx %6.2f %2sps Tx %5llu pps Rx %5llu pps Tx\n",
		       ifname,
		       rxround,
		       rxsuffix,
		       txround,
		       txsuffix,
		       (unsigned long long)rxpps,
		       (unsigned long long)txpps);
	} else {
		printf("%llu %llu %llu %llu\n",
		       (unsigned long long)rxbits,
		       (unsigned long long)txbits,
		       (unsigned long long)rxpps,
		       (unsigned long long)txpps);
	}
	fflush(stdout);
}

void
printrxtx(int sig)
{
	char *rxsuffix, *txsuffix;
	double rxround, txround;
	uint64_t rxbits = (rxbytes - origrxbytes) * 8,
	         txbits = (txbytes - origtxbytes) * 8;

	rxround = scale(&rxsuffix, rxbits);
	txround = scale(&txsuffix, txbits);
	printf("%*s %6.2f %2s   Rx %6.2f %2s   Tx\n",
	       (int)strlen(ifname) + 1, "",
	       rxround, rxsuffix, txround, txsuffix);
}

int
getmonotime(struct timeval *tv)
{
	struct timespec ts;
	int ret;

	ret = clock_gettime(CLOCK_MONOTONIC, &ts);
	if (ret < 0)
		return -1;
	tv->tv_sec = ts.tv_sec;
	tv->tv_usec = ts.tv_nsec / 1000;
	return ret;
}

void
ms2tv(struct timeval *tv, long ms)
{
	tv->tv_sec = ms / 1000;
	tv->tv_usec = (ms % 1000) * 1000;
}

long
tv2ms(struct timeval *tv)
{
	return (tv->tv_sec * 1000) + ((tv->tv_usec + 500) / 1000);
}

int
tvsleep(struct timeval *tv)
{
	struct timespec ts;

	ts.tv_sec = tv->tv_sec;
	ts.tv_nsec = tv->tv_usec * 1000;
	return nanosleep(&ts, NULL);
}

void
loop(long count, struct timeval *delay)
{
	struct timeval old, now, diff;
	uint64_t oldrxbytes, diffrxbits;
	uint64_t oldtxbytes, difftxbits;
	uint64_t oldrxpps, rxpps, diffrxpps;
	uint64_t oldtxpps, txpps, difftxpps;
	long n = 0;

	getmonotime(&old);
	sample(&oldrxbytes, &oldtxbytes, &oldrxpps, &oldtxpps);
	origrxbytes = oldrxbytes, origtxbytes = oldtxbytes;
	for (;;) {
		tvsleep(delay);
		sample(&rxbytes, &txbytes, &rxpps, &txpps);
		getmonotime(&now);
		timersub(&now, &old, &diff);
		getmonotime(&old);
		diffrxbits = (rxbytes - oldrxbytes) * 8;
		difftxbits = (txbytes - oldtxbytes) * 8;
		diffrxbits = diffrxbits * 1000 / tv2ms(&diff);
		difftxbits = difftxbits * 1000 / tv2ms(&diff);
		diffrxpps = (rxpps - oldrxpps) * 1000 / tv2ms(&diff);
		difftxpps = (txpps - oldtxpps) * 1000 / tv2ms(&diff);
		print(diffrxbits, difftxbits, diffrxpps, difftxpps);
		oldrxbytes = rxbytes, oldtxbytes = txbytes;
		oldrxpps = rxpps, oldtxpps = txpps;
		if (count && ++n >= count)
			break;
	}
}

void
usage(void)
{
	fprintf(stderr, "usage: %s [-B | -t] [-c count] [-d delay] [-i interface]\n", argv0);
	exit(1);
}

int
main(int argc, char *argv[])
{
	char ifn[IF_NAMESIZE] = "";
	char *end;
	struct timeval tv;
	long count = 0, delay = 1000;

	ARGBEGIN {
	case 'B':
		Bflag = 1;
		break;
	case 'c':
		errno = 0;
		count = strtol(EARGF(usage()), &end, 10);
		if (*end != '\0' || errno)
			errx(1, "invalid count");
		break;
	case 'd':
		errno = 0;
		delay = strtol(EARGF(usage()), &end, 10);
		if (*end != '\0' || errno)
			errx(1, "invalid delay");
		break;
	case 'i':
		strncpy(ifn, EARGF(usage()), sizeof(ifn));
		ifn[IF_NAMESIZE - 1] = '\0';
		break;
	case 't':
		tflag = 1;
		break;
	default:
		usage();
	} ARGEND

	ifname = ifn;
	signal(SIGINFO, printrxtx);
	ms2tv(&tv, delay);
	scan();
	loop(count, &tv);
	return 0;
}
